import importlib.resources
import re
from pathlib import Path
from typing import Any, List, Literal, Optional, Union

import pytz
import yaml
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    FilePath,
    ValidationError,
    field_validator,
)
from pydantic_core.core_schema import ValidationInfo
from pytz.exceptions import UnknownTimeZoneError
from typing_extensions import Annotated

from focus_converter.conversion_functions import STATIC_CONVERSION_TYPES
from focus_converter.models.focus_column_names import FocusColumnNames


class SQLConditionConversionArgs(BaseModel):
    conditions: List[str]
    default_value: Any


class LookupConversionArgs(BaseModel):
    reference_path_in_package: bool = True
    reference_dataset_path: FilePath
    source_value: str
    destination_value: str

    @field_validator("reference_dataset_path", mode="before")
    def __validate_reference_dataset_path__(
        cls, reference_dataset_path, field_info: ValidationInfo
    ):
        reference_path_in_package = field_info.data.get("reference_path_in_package")
        if reference_path_in_package:
            try:
                return (
                    importlib.resources.files("focus_converter")
                    .joinpath(reference_dataset_path)
                    .as_posix()
                )
            except AttributeError:
                # for older python versions wheres .files api is not available
                from pkg_resources import resource_filename

                return resource_filename("focus_converter", reference_dataset_path)
        else:
            return reference_dataset_path


class ValueMapItemConversionArgs(BaseModel):
    key: str
    value: str


class ValueMapConversionArgs(BaseModel):
    value_list: List[ValueMapItemConversionArgs]
    default_value: str

    # flag decides if default value should be applied on null values or leave it null if false
    apply_default_if_null: Optional[bool] = True


class StaticValueConversionArgs(BaseModel):
    static_value: Any


class UnnestValueConversionArgs(BaseModel):
    # child element type, could be a struct or a list of structs
    children_type: Literal["list", "struct"] = "struct"

    # default behaviour to return first value, since we don't know what is the datatype
    aggregation_operation: Optional[
        Literal["first", "last", "sum", "mean", "min", "max"]
    ] = "first"


class MissingColumnDType(BaseModel):
    data_type: Literal["string", "float", "int"]


class DTypeConversionArg(BaseModel):
    column_name: str
    dtype: Literal["string", "float", "int"]
    strict: bool = False


class SetColumnDTypesConversionArgs(BaseModel):
    dtype_args: List[DTypeConversionArg]


CONFIG_FILE_PATTERN = re.compile("(.+)_S[0-9]{3}.yaml")


class ConversionPlan(BaseModel):
    # config source file name
    config_file_name: str

    # a friendly name helpful for debugging purposes
    plan_name: str

    # id picked from the file name, starting with D.
    dimension_id: Union[str, int]

    # for some datasets we might have to do conversions in an order,
    # higher priority will guarantee that those plans are executed first.
    priority: int

    # column to apply this on
    # TODO: Add option to allow query from multiple columns
    column: str

    conversion_type: STATIC_CONVERSION_TYPES
    conversion_args: Annotated[Any, Field(validate_default=True)] = None

    # converted focus column name
    focus_column: FocusColumnNames

    # optional column fix useful for intermediate columns to be dropped from final dataset
    # this is optional, but if defined should have the suffix `tmp`
    column_prefix: Optional[str] = None

    @field_validator("conversion_args")
    @classmethod
    def conversion_args_validation(cls, v: Any, field_info: ValidationInfo) -> str:
        conversion_type: STATIC_CONVERSION_TYPES = field_info.data.get(
            "conversion_type"
        )

        if (
            conversion_type == STATIC_CONVERSION_TYPES.ASSIGN_TIMEZONE
            or conversion_type == STATIC_CONVERSION_TYPES.CONVERT_TIMEZONE
        ):
            try:
                pytz.timezone(v)
            except UnknownTimeZoneError:
                raise ValueError(
                    f"Invalid timezone specified in conversion_args for plan: {field_info.data}"
                )
        elif conversion_type == STATIC_CONVERSION_TYPES.SQL_CONDITION:
            try:
                SQLConditionConversionArgs.model_validate(v)
            except ValidationError:
                raise ValueError(
                    f"Invalid SQL condition specified in conversion_args for plan: {field_info.data}"
                )
        elif conversion_type == STATIC_CONVERSION_TYPES.LOOKUP:
            try:
                LookupConversionArgs.model_validate(v)
            except ValidationError as e:
                raise ValueError(
                    e,
                    f"Invalid lookup arg specified in conversion_args for plan: {field_info.data}",
                )
        elif conversion_type == STATIC_CONVERSION_TYPES.MAP_VALUES:
            try:
                ValueMapConversionArgs.model_validate(v)
            except ValidationError as e:
                raise ValueError(
                    e, f"Missing or bad mapping value argument: {field_info.data}"
                )
        elif conversion_type == STATIC_CONVERSION_TYPES.ASSIGN_STATIC_VALUE:
            try:
                StaticValueConversionArgs.model_validate(v)
            except ValidationError as e:
                raise ValueError(
                    e, f"Missing or bad static value argument: {field_info.data}"
                )
        elif conversion_type == STATIC_CONVERSION_TYPES.APPLY_DEFAULT_IF_COLUMN_MISSING:
            try:
                MissingColumnDType.model_validate(v)
            except ValidationError as e:
                raise ValueError(
                    e, f"Missing or bad unnest value argument: {field_info.data}"
                )
        elif conversion_type == STATIC_CONVERSION_TYPES.SET_COLUMN_DTYPES:
            try:
                SetColumnDTypesConversionArgs.model_validate(v)
            except ValidationError as e:
                raise ValueError(
                    e, f"Missing or bad set column dtype argument: {field_info.data}"
                )
        return v

    @field_validator("column_prefix")
    def validate_focus_column(cls, v: Any):
        if isinstance(v, str) and not v.startswith("tmp_"):
            raise ValueError("column_prefix, if defined should have 'tmp_' prefix.")
        return v

    @staticmethod
    def load_yaml(config_path: FilePath):
        if isinstance(config_path, str):
            config_path = Path(config_path)

        with open(config_path) as fd:
            obj = yaml.safe_load(fd)
        assert (
            re.match(CONFIG_FILE_PATTERN, config_path.name) is not None
        ), f"Filename should match pattern: {CONFIG_FILE_PATTERN}"

        # extract priority of config from name starting with S in the pattern D000_S000.yaml
        dimension_id = re.search(r"(.+)_S\d+\.yaml", config_path.name).group(1)
        priority = re.search(r".+_S(\d+)\.yaml", config_path.name).group(1)
        obj["priority"] = int(priority)
        obj["dimension_id"] = dimension_id
        obj["config_file_name"] = config_path.name

        return ConversionPlan.model_validate(obj=obj)

    model_config = ConfigDict(extra="forbid")



import polars as pl

from focus_converter.configs.base_config import ConversionPlan
from focus_converter.conversion_functions.validations import ColumnValidator


class DateTimeConversionFunctions:
    @staticmethod
    def convert_timezone(
        plan: ConversionPlan, column_alias, column_validator: ColumnValidator
    ) -> pl.col:
        # add to column validator and check if source column exists
        column_validator.map_non_sql_plan(plan=plan, column_alias=column_alias)

        return (
            pl.col(plan.column)
            .dt.cast_time_unit("ms")
            .dt.convert_time_zone(plan.conversion_args)
            .alias(column_alias)
        )

    @staticmethod
    def assign_timezone(
        plan: ConversionPlan, column_alias, column_validator: ColumnValidator
    ) -> pl.col:
        # add to column validator and check if source column exists
        column_validator.map_non_sql_plan(plan=plan, column_alias=column_alias)

        return (
            pl.col(plan.column)
            .dt.cast_time_unit("ms")
            .dt.replace_time_zone(plan.conversion_args, ambiguous="earliest")
            .alias(column_alias)
        )

    @staticmethod
    def assign_utc_timezone(
        plan: ConversionPlan, column_alias, column_validator: ColumnValidator
    ) -> pl.col:
        # add to column validator and check if source column exists
        column_validator.map_non_sql_plan(plan=plan, column_alias=column_alias)

        return (
            pl.col(plan.column)
            .dt.cast_time_unit("ms")
            .dt.replace_time_zone("UTC")
            .alias(column_alias)
        )

    @staticmethod
    def parse_datetime(
        plan: ConversionPlan, column_alias, column_validator: ColumnValidator
    ) -> pl.col:
        # add to column validator and check if source column exists
        column_validator.map_non_sql_plan(plan=plan, column_alias=column_alias)

        return (
            pl.col(plan.column)
            .str.strptime(pl.Datetime, plan.conversion_args)
            .dt.cast_time_unit("ms")
            .alias(column_alias)
        )



from enum import Enum

DEFAULT_FOCUS_NAMESPACE = "F"


class FocusColumnNames(Enum):
    """
    Focus column names as described in https://focus.finops.org/#specification
    """

    PLACE_HOLDER = "PlaceHolder"

    CHARGE_PERIOD_START = "ChargePeriodStart"
    CHARGE_PERIOD_END = "ChargePeriodEnd"
    CHARGE_FREQUENCY = "ChargeFrequency"
    CHARGE_SUB_CATEGORY = "ChargeSubcategory"
    CHARGE_DESCRIPTION = "ChargeDescription"

    BILLING_PERIOD_START = "BillingPeriodStart"
    BILLING_PERIOD_END = "BillingPeriodEnd"

    PROVIDER = "Provider"
    PUBLISHER = "Publisher"
    INVOICE_ISSUER = "InvoiceIssuer"

    BILLING_ACCOUNT_ID = "BillingAccountId"
    BILLING_ACCOUNT_NAME = "BillingAccountName"

    AMORTISED_COST = "AmortisedCost"
    BILLED_COST = "BilledCost"
    BILLED_CURRENCY = "BilledCurrency"
    BILLING_CURRENCY = "BillingCurrency"
    EFFECTIVE_COST = "EffectiveCost"

    REGION = "Region"

    CHARGE_TYPE = "ChargeType"

    SERVICE_CATEGORY = "ServiceCategory"
    SERVICE_NAME = "ServiceName"

    SUB_ACCOUNT_NAME = "SubAccountName"
    SUB_ACCOUNT_ID = "SubAccountId"

    AVAILABILITY_ZONE = "AvailabilityZone"

    RESOURCE_NAME = "ResourceName"
    RESOURCE_ID = "ResourceId"

    COMMITMENT_DISCOUNT_ID = "CommitmentDiscountId"
    COMMITMENT_DISCOUNT_TYPE = "CommitmentDiscountType"
    COMMITMENT_DISCOUNT_CATEGORY = "CommitmentDiscountCategory"

    PRICING_QUANTITY = "PricingQuantity"
    PRICING_UNIT = "PricingUnit"

    SKU_ID = "SkuId"
    SKU_PRICE_ID = "SkuPriceId"

    LIST_UNIT_PRICE = "ListUnitPrice"
    LIST_COST = "ListCost"

    USAGE_QUANTITY = "UsageQuantity"
    USAGE_UNIT = "UsageUnit"


FOCUS_DATETIME_ISO_FORMAT = "%Y-%m-%dT%H:%M:%SZ"



from enum import Enum
from typing import Iterable

import pandas as pd
import polars as pl
import pyarrow.dataset as ds
from tqdm import tqdm

# these values need to be tweaked, for datasets with small number of columns,
# following values can be much larger.
# for very fragments with large number of row groups, this will cause memory to exhaust
DEFAULT_BATCH_READ_SIZE = 50000
FRAGMENT_READ_AHEAD = 0
BATCH_READ_AHEAD = 0


class DataFormats(Enum):
    CSV = "csv"
    PARQUET = "parquet"


class ParquetDataFormat(Enum):
    FILE = "file"
    DATASET = "dataset"
    DELTA = "delta"


class DataLoader:
    def __init__(
        self,
        data_path: str,
        data_format: DataFormats,
        parquet_data_format: ParquetDataFormat,
    ):
        self.__data_path__ = data_path
        self.__data_format__ = data_format
        self.__parquet_data_format__ = parquet_data_format

    def load_pyarrow_dataset(self) -> Iterable[pl.LazyFrame]:
        dataset = ds.dataset(self.__data_path__)
        scanner = dataset.scanner(
            batch_size=DEFAULT_BATCH_READ_SIZE,
            use_threads=True,
            fragment_readahead=FRAGMENT_READ_AHEAD,
            batch_readahead=BATCH_READ_AHEAD,
        )

        total_rows = dataset.count_rows()

        with tqdm(total=total_rows) as pobj:
            for batch in scanner.to_batches():
                df = pl.from_arrow(batch)
                yield df.lazy()
                pobj.update(df.shape[0])

    def load_parquet_file(self) -> Iterable[pl.LazyFrame]:
        # reads parquet from data path and returns a lazy object

        yield pl.read_parquet(self.__data_path__).lazy()

    def load_csv(self) -> Iterable[pl.LazyFrame]:
        # reads csv from data path and returns a lazy object

        yield pl.read_csv(
            self.__data_path__, try_parse_dates=True, ignore_errors=True
        ).lazy()

    def data_scanner(self) -> Iterable[pl.LazyFrame]:
        # helper function to read from different data formats and create an iterator of lazy frames
        # which then can be used to apply lazy eval plans

        if self.__data_format__ == DataFormats.CSV:
            yield from self.load_csv()
        elif self.__data_format__ == DataFormats.PARQUET:
            if self.__parquet_data_format__ == ParquetDataFormat.FILE:
                yield from self.load_parquet_file()
            elif self.__parquet_data_format__ == ParquetDataFormat.DATASET:
                yield from self.load_pyarrow_dataset()
            else:
                raise NotImplementedError(
                    f"Parquet format:{self.__parquet_data_format__} not implemented"
                )
        else:
            raise NotImplementedError(
                f"Data format:{self.__data_format__} not implemented"
            )



import base64
import io

import networkx as nx
import polars as pl
import requests
import sqlglot

from focus_converter.configs.base_config import (
    ConversionPlan,
    SetColumnDTypesConversionArgs,
)

SOURCE_COLUMN_NAME = "SOURCE"
SINK_COLUMN_NAME = "FOCUS_DATASET"
STATIC_VALUE_COLUMN = "STATIC_VALUE"


def mm(graph):
    # generates UML using mermaid's public api
    # TODO: Find a robust local graph draw

    graphbytes = graph.encode("ascii")
    base64_bytes = base64.b64encode(graphbytes)
    base64_string = base64_bytes.decode("ascii")
    return requests.get(f"https://mermaid.ink/img/{base64_string}").content


class ColumnValidator:
    def __init__(self):
        self.__network_graph__ = network_graph = nx.DiGraph()
        network_graph.add_node(SOURCE_COLUMN_NAME)
        network_graph.add_node(SINK_COLUMN_NAME)

    def __validate_column_names__(self, plan: ConversionPlan, column_names):
        # ensures that dependent columns are present in the data

        for column_name in column_names:
            if (
                column_name not in self.__network_graph__.nodes
                and column_name != STATIC_VALUE_COLUMN
            ):
                # add edge indicating that these are to be provided from source data
                # if the edge not present, then it is expected to come from the source dataset
                self.__network_graph__.add_edge(SOURCE_COLUMN_NAME, column_name)

    def __add_sink_node__(self, focus_column):
        self.__network_graph__.add_edge(focus_column, SINK_COLUMN_NAME)

    def map_sql_query(self, sql_query, plan: ConversionPlan):
        """
        Maps a sql query to the graph

        :param sql_query: str, sql query to be mapped
        :param plan: ConversionPlan, plan to be added to the graph
        """

        # generate a list of all columns that are aliased
        source_columns = sorted(
            set(
                [
                    column.alias_or_name
                    for column in sqlglot.parse_one(sql_query).find_all(
                        sqlglot.exp.Column
                    )
                ]
            )
        )
        self.__validate_column_names__(plan=plan, column_names=source_columns)

        # generate sorted list of target columns from the query
        target_columns = sorted(
            set(
                [
                    column.alias_or_name
                    for column in sqlglot.parse_one(sql_query).find_all(
                        sqlglot.exp.Alias
                    )
                ]
            )
        )

        for source_column in source_columns:
            for target_column in target_columns:
                self.__network_graph__.add_edge(source_column, target_column, plan=plan)

        # analyze target_columns that are focus columns and automatically add sink node
        for target_column in target_columns:
            if target_column == plan.focus_column.value:
                self.__add_sink_node__(focus_column=target_column)

    def map_non_sql_plan(
        self, plan: ConversionPlan, column_alias, source_column: str = None
    ):
        """
        Maps a non sql plan to the graph

        :param plan: ConversionPlan, plan to be added to the graph
        :param column_alias: str, alias to be used for the column
        :param source_column: str: Optional source column name if different from plan.column, used for unnest
        :return:
        """

        self.__validate_column_names__(
            plan=plan, column_names=[source_column or plan.column]
        )
        self.__network_graph__.add_edge(
            source_column or plan.column, column_alias, plan=plan
        )
        self.__add_sink_node__(focus_column=plan.focus_column.value)

    def map_static_default_value_if_not_present(
        self, plan: ConversionPlan, column_alias
    ):
        self.__network_graph__.add_edge(SOURCE_COLUMN_NAME, plan.column, plan=plan)

    def map_dtype_enforced_node(self, plan: ConversionPlan):
        conversion_args = SetColumnDTypesConversionArgs.model_validate(
            plan.conversion_args
        )
        for column_obj in conversion_args.dtype_args:
            self.__network_graph__.add_edge(
                SOURCE_COLUMN_NAME, column_obj.column_name, plan=plan
            )

    def validate_lazy_frame_columns(self, lf: pl.LazyFrame):
        # get all columns that have edge from source
        source_columns = [
            column for column in self.__network_graph__.successors(SOURCE_COLUMN_NAME)
        ]

        columns_missing = sorted(set(source_columns) - set(lf.columns))
        if columns_missing:
            raise ValueError(
                f"Column(s) '{', '.join(columns_missing)}' not found in data"
            )

    def validate_graph_is_connected(self):
        """
        Validates that the graph is connected
        """

        graph = self.__network_graph__

        sink_nodes = [node for node in graph.nodes() if graph.out_degree(node) == 0]
        sink_nodes = set(sink_nodes) - {SOURCE_COLUMN_NAME, SINK_COLUMN_NAME}

        if len(sink_nodes) > 0:
            raise ValueError(
                "Following sink nodes are not connected, potentially missing transform steps",
                sink_nodes,
            )

    def generate_mermaid_uml(self):
        """
        Using networkx, generate a mermaid compatible UML graph of the conversion plan
        """

        graph_uml = io.StringIO()
        graph_uml.write("graph TD;\n")

        for source, target, edge_data in self.__network_graph__.edges(data=True):
            plan: ConversionPlan = edge_data.get("plan")
            if plan:
                graph_uml.write(
                    f"\t{source} -- {plan.conversion_type.name}:{plan.config_file_name} --> {target}\n"
                )
            else:
                graph_uml.write(f"\t{source} --> {target}\n")

        graph_uml.seek(0)
        return graph_uml.read()

    def generate_uml_graph(self):
        mermaid_graph = self.generate_mermaid_uml()
        return mm(mermaid_graph)



import importlib.resources
import logging
import os
from operator import attrgetter
from typing import Dict, List, Optional

import polars as pl

from focus_converter.configs.base_config import ConversionPlan
from focus_converter.conversion_functions import STATIC_CONVERSION_TYPES
from focus_converter.conversion_functions.column_functions import ColumnFunctions
from focus_converter.conversion_functions.datetime_functions import (
    DateTimeConversionFunctions,
)
from focus_converter.conversion_functions.deferred_column_functions import (
    DeferredColumnFunctions,
)
from focus_converter.conversion_functions.lookup_function import LookupFunction
from focus_converter.conversion_functions.sql_functions import SQLFunctions
from focus_converter.conversion_functions.validations import ColumnValidator
from focus_converter.data_loaders.data_exporter import DataExporter
from focus_converter.data_loaders.data_loader import DataLoader
from focus_converter.models.focus_column_names import FocusColumnNames

# TODO: Make this path configurable so that we can load from a directory outside of the project
BASE_CONVERSION_CONFIGS = (
    importlib.resources.files("focus_converter")
    .joinpath("conversion_configs")
    .as_posix()
)


class FocusConverter:
    plans: Dict[str, List[ConversionPlan]]
    data_loader: DataLoader
    data_exporter: DataExporter = None

    # set of plan variables for horizontal transformation plans
    h_collected_columns: List[str]  # collected columns
    h_column_exprs: List[pl.col]  # column expressions
    h_sql_queries: List[str]  # sql_queries

    # temporary columns to be removed from final dataset
    __temporary_columns__: List[str]

    # column prefix used in source dataset
    __column_prefix__: Optional[str] = None

    # converted column prefix to be added to converted columns
    __converted_column_prefix__: Optional[str] = None

    # deferred column plans, these plans are applied after lazyframe is loaded
    __deferred_column_plans__ = DeferredColumnFunctions()

    def __init__(self, column_prefix=None, converted_column_prefix=None):
        self.__temporary_columns__ = []
        self.__column_prefix__ = column_prefix
        self.__converted_column_prefix__ = converted_column_prefix

        # ColumnValidator, used to validate column names in sql queries and transformations
        self.__column_validator__ = ColumnValidator()

    def load_provider_conversion_configs(self):
        plans = {}

        for provider in os.listdir(BASE_CONVERSION_CONFIGS):
            # collect all plans specific to vendor
            provider_plans: List[ConversionPlan] = []

            provider_base_path = os.path.join(BASE_CONVERSION_CONFIGS, provider)
            for provider_config_name in os.listdir(provider_base_path):
                if not provider_config_name.endswith(".yaml"):
                    # ignores non yaml files, which may be reference datasets
                    continue

                provider_config_path = os.path.join(
                    provider_base_path, provider_config_name
                )
                logging.debug(f"Reading plan from path: {provider_config_name}")
                provider_plans.append(ConversionPlan.load_yaml(provider_config_path))

            plans[provider] = sorted(
                provider_plans,
                key=attrgetter("dimension_id", "priority"),
                reverse=False,
            )
        self.plans = plans

    def load_data(self, *args, **kwargs):
        self.data_loader = DataLoader(*args, **kwargs)

    def configure_data_export(self, *args, **kwargs):
        self.data_exporter = DataExporter(*args, **kwargs)

    def prepare_horizontal_conversion_plan(self, provider):
        # final set of columns produced after this transform step
        self.h_collected_columns = collected_columns = []

        # column expressions that will be applied to loaded lazy frame in order
        self.h_column_exprs = column_exprs = []

        # sql queries collected to be applied on the lazy frame
        self.h_sql_queries = sql_queries = []

        # lookup lazyframes arguments to be assembled later on the final source lazyframe
        self.lookup_reference_args = []

        for plan in self.plans[provider]:
            # column name generated with temporary prefix
            if plan.column_prefix:
                column_alias = f"{plan.column_prefix}_{plan.focus_column.value}"
                self.__temporary_columns__.append(column_alias)
            else:
                column_alias = plan.focus_column.value

            # add column to plan to collect these dimensions to be added in the computed dataframe
            if plan.focus_column != FocusColumnNames.PLACE_HOLDER:
                collected_columns.append(plan.focus_column.value)

            if plan.conversion_type == STATIC_CONVERSION_TYPES.CONVERT_TIMEZONE:
                column_exprs.append(
                    DateTimeConversionFunctions.convert_timezone(
                        plan=plan,
                        column_alias=column_alias,
                        column_validator=self.__column_validator__,
                    )
                )
            elif plan.conversion_type == STATIC_CONVERSION_TYPES.ASSIGN_TIMEZONE:
                column_exprs.append(
                    DateTimeConversionFunctions.assign_timezone(
                        plan=plan,
                        column_alias=column_alias,
                        column_validator=self.__column_validator__,
                    )
                )
            elif plan.conversion_type == STATIC_CONVERSION_TYPES.ASSIGN_UTC_TIMEZONE:
                column_exprs.append(
                    DateTimeConversionFunctions.assign_utc_timezone(
                        plan=plan,
                        column_alias=column_alias,
                        column_validator=self.__column_validator__,
                    )
                )
            elif plan.conversion_type == STATIC_CONVERSION_TYPES.RENAME_COLUMN:
                column_exprs.append(
                    ColumnFunctions.rename_column_functions(
                        plan=plan,
                        column_alias=column_alias,
                        column_validator=self.__column_validator__,
                    )
                )
            elif plan.conversion_type == STATIC_CONVERSION_TYPES.SQL_QUERY:
                sql_queries.append(
                    SQLFunctions.eval_sql_query(
                        plan=plan,
                        column_alias=column_alias,
                        column_validator=self.__column_validator__,
                    )
                )
            elif plan.conversion_type == STATIC_CONVERSION_TYPES.SQL_CONDITION:
                sql_queries.append(
                    SQLFunctions.eval_sql_conditions(
                        plan=plan,
                        column_alias=column_alias,
                        column_validator=self.__column_validator__,
                    )
                )
            elif plan.conversion_type == STATIC_CONVERSION_TYPES.PARSE_DATETIME:
                column_exprs.append(
                    DateTimeConversionFunctions.parse_datetime(
                        plan=plan,
                        column_alias=column_alias,
                        column_validator=self.__column_validator__,
                    )
                )
            elif plan.conversion_type == STATIC_CONVERSION_TYPES.UNNEST_COLUMN:
                column_exprs.append(
                    ColumnFunctions.unnest(
                        plan=plan,
                        column_alias=column_alias,
                        column_validator=self.__column_validator__,
                    )
                )
            elif plan.conversion_type == STATIC_CONVERSION_TYPES.LOOKUP:
                self.lookup_reference_args.append(
                    LookupFunction.map_values_using_lookup(
                        plan=plan,
                        column_alias=column_alias,
                        column_validator=self.__column_validator__,
                    )
                )
            elif plan.conversion_type == STATIC_CONVERSION_TYPES.MAP_VALUES:
                column_exprs.append(
                    ColumnFunctions.map_values(
                        plan=plan,
                        column_alias=column_alias,
                        column_validator=self.__column_validator__,
                    )
                )
            elif plan.conversion_type == STATIC_CONVERSION_TYPES.ASSIGN_STATIC_VALUE:
                column_exprs.append(
                    ColumnFunctions.assign_static_value(
                        plan=plan,
                        column_alias=column_alias,
                        column_validator=self.__column_validator__,
                    )
                )
            elif (
                plan.conversion_type
                == STATIC_CONVERSION_TYPES.APPLY_DEFAULT_IF_COLUMN_MISSING
            ):
                self.__deferred_column_plans__.map_missing_column_plan(
                    plan=plan,
                    column_alias=column_alias,
                    column_validator=self.__column_validator__,
                )
            elif plan.conversion_type == STATIC_CONVERSION_TYPES.SET_COLUMN_DTYPES:
                self.__deferred_column_plans__.map_dtype_plan(
                    plan=plan, column_validator=self.__column_validator__
                )
            else:
                raise NotImplementedError(
                    f"Plan: {plan.conversion_type} not implemented"
                )

        # apply the plan to the lazy frame
        self.__column_validator__.validate_graph_is_connected()

        return column_exprs

    @staticmethod
    def __apply_sql_queries__(lf: pl.LazyFrame, sql_queries):
        for sql_query in sql_queries:
            sql_context = SQLFunctions.create_sql_context(lf=lf)
            lf = sql_context.execute(sql_query, eager=False)
        return lf

    @staticmethod
    def __apply_column_expressions__(
        lf: pl.LazyFrame, column_expressions: List[pl.col]
    ):
        for expr in column_expressions:
            lf = lf.with_columns_seq(expr)
        return lf

    @staticmethod
    def __apply_lookup_reference_plans__(lf: pl.LazyFrame, lookup_args):
        for lookup_arg in lookup_args:
            lf = lf.join(**lookup_arg)
        return lf

    def explain(self):
        # get batched data lazy frame, build the plan and then break
        return self.__column_validator__.generate_uml_graph()

    def apply_plan(self, lf: pl.LazyFrame) -> pl.LazyFrame:
        # creates lazy frame using the config, actual computation happens in collect

        lf = self.__apply_column_expressions__(
            lf=lf, column_expressions=self.h_column_exprs
        )

        # apply lazyframe joins
        lf = self.__apply_lookup_reference_plans__(
            lf=lf, lookup_args=self.lookup_reference_args
        )

        lf = self.__apply_sql_queries__(lf=lf, sql_queries=self.h_sql_queries)

        # drop temporary columns from the final dataset
        lf = lf.drop(self.__temporary_columns__)

        return lf

    def __re_map_source_columns__(self, lf: pl.LazyFrame):
        # helper function to re-map prefixed source columns to source columns
        # so that conversion plans can be applied.

        column_prefix_length = len(self.__column_prefix__)
        for column in lf.columns:
            if column.startswith(self.__column_prefix__):
                orig_column_name = column[column_prefix_length:]
                lf = lf.with_columns_seq(pl.col(column).alias(orig_column_name))

                # add remapped column to temporary columns list so that it can be dropped later
                self.__temporary_columns__.append(orig_column_name)
        return lf

    def __process_lazy_frame__(self, lf: pl.LazyFrame):
        # prepares lazyframe for the operations to be applied on the lazy loaded polars dataframe
        if self.__column_prefix__ is not None:
            lf = self.__re_map_source_columns__(lf=lf)

        # apply deferred column plans
        lf = self.__deferred_column_plans__.apply_missing_column_plan(lf=lf)
        lf = self.__deferred_column_plans__.apply_dtype_plan(lf=lf)

        # validate all source columns exist in the lazy frame
        self.__column_validator__.validate_lazy_frame_columns(lf=lf)

        return self.apply_plan(lf=lf)

    def convert(self):
        error = None

        for lf in self.data_loader.data_scanner():
            try:
                lf = self.__process_lazy_frame__(lf=lf)
                self.data_exporter.collect(
                    lf=lf, collected_columns=list(set(self.h_collected_columns))
                )
            except Exception as e:
                error = e
                break

        if error is not None:
            # cleanup data export now that we have to close all subprocesses
            if self.data_exporter:
                self.data_exporter.close()

            raise error
